#include <stdio.h>
#include <unistd.h>
#include <sys/pstat.h>
#include <stdlib.h>
#include <pwd.h>
#include <string.h>

static void print_process(const struct pst_status *pst, int indent) {
	printf("%*s%ld %s\n", indent*2, "", (long)pst->pst_pid, pst->pst_cmd);
}


static void traverse_ancestors(pid_t child_pid, int *indent) {
	struct pst_status pst;
	int rc = pstat_getproc(&pst, sizeof(pst), 0, child_pid);
	if(rc<1) {
		perror("pstat");
		exit(1);
	}
	if(pst.pst_ppid==1) {
		//Got the init process (which we never print)
		*indent = 0;
	} else {
		if(pst.pst_ppid!=pst.pst_pid)
			traverse_ancestors(pst.pst_ppid,indent);
		print_process(&pst,*indent);
		++*indent;
	}
}


static void traverse_children(pid_t parent_pid, int indent) {
	struct pst_status pst[64];
	struct pst_status *pst_child = 0;
	int children = 0;
	int proc_index=0;
	int rc;
	while((rc=pstat_getproc(pst,sizeof(pst[0]), 64, proc_index))>0 ) {
		for(int i=0; i<rc; i++) {
			if(pst[i].pst_ppid==parent_pid) {
				int is_duplicate = 0;
				for(int j=0; j<children; j++)
					if(pst_child[j].pst_pid==pst[i].pst_pid)
						is_duplicate = 1;
				if(!is_duplicate) {
					pst_child = (struct pst_status*)realloc(pst_child, (children+1)*sizeof(pst_child[0]));
					pst_child[children++] = pst[i];
				}
			}
		}
		proc_index = pst[rc-1].pst_idx+1;
	}
	for(int i=0; i<children; i++) {
		print_process(pst_child+i,indent+1);
		if(pst_child[i].pst_pid!=pst_child[i].pst_ppid)
			traverse_children(pst_child[i].pst_pid,indent+1);
	}
	free(pst_child);
}



static void show_tree_by_pid(pid_t goal_pid) {
	int indent=0;
	traverse_ancestors(goal_pid,&indent);
	//print_process(goal_pid,indent);
	traverse_children(goal_pid,indent-1);
}


static void show_subtree_by_uid(struct pst_status *pst, size_t psts, int root_index, int *shown, int indent) {
	print_process(pst+root_index, indent);
	shown[root_index] = 1;
	for(size_t i=0; i<psts; i++)
		if(pst[i].pst_ppid==pst[root_index].pst_pid)
			show_subtree_by_uid(pst,psts, i, shown, indent+1);
}


static void show_tree_by_uid(uid_t uid) {
	struct pst_status *pst = 0;
	size_t psts = 0;
	int proc_index=0;
	
	//grab the whole process list
	for(;;) {
		pst = (struct pst_status*)realloc(pst, sizeof(*pst)*(psts+64));
		int rc = pstat_getproc(pst+psts, sizeof(*pst), 64, proc_index);
		if(rc<=0)
			break;
		psts += rc;
		proc_index = pst[psts-1].pst_idx+1;
	}
	
	//eleminate processes not belonging to the user
	{
		size_t dst = 0;
		for(size_t src=0; src<psts; src++) {
			if(pst[src].pst_uid==uid) {
				if(src!=dst)
					memcpy(pst+dst, pst+src, sizeof(*pst));
				dst++;
			}
		}
		psts = dst;
	}
	
	//show all the (sub-)trees
	int *shown = (int*)calloc(psts, sizeof(int));
	for(size_t i=0; i<psts; i++) {
		if(shown[i])
			continue;
		//is it a root?
		int is_a_root=1;
		for(size_t j=0; j<psts; j++)
			if(pst[j].pst_pid == pst[i].pst_ppid)
				is_a_root=0;
		if(is_a_root) {
			show_subtree_by_uid(pst,psts,i,shown,0);
		}
	}
	free(shown);
	free(pst);
}


int main(int argc, char **argv) {
	if(argc>2 ||
	  (argc==2 && (strcmp(argv[1],"-?")==0 ||
	               strcmp(argv[1],"-usage")==0 ||
	               strcmp(argv[1],"--usage")==0 ||
	               strcmp(argv[1],"--help")==0)))
	{
		fprintf(stderr,"usage: ptree [pid|uid]\n");
		fprintf(stderr,"  (show process trees)\n");
		return 99;
	}
	if(argc==1) {
		show_tree_by_pid(1);
	} else {
		char *endptr=0;
		long goal_pid = strtol(argv[1],&endptr,0);
		if(endptr && *endptr) {
			//must be a username
			struct passwd *pw = getpwnam(argv[1]);
			if(!pw) {
				fprintf(stderr,"ptree: User '%s' not found\n",argv[1]);
				return 2;
			}
			show_tree_by_uid(pw->pw_uid);
		} else {
			//show tree by process-id
			show_tree_by_pid(goal_pid);
		}
	}
	return 0;
}
